"""Azure IoT Hub Device SDK Models

This package provides object models for use within the Azure IoT Hub Device SDK.
"""

from .message import Message
from .methods import MethodRequest, MethodResponse



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import sys
from azure.iot.device.common.evented_callback import EventedCallback
from azure.iot.device.common.pipeline import (
    pipeline_stages_base,
    pipeline_ops_base,
    pipeline_stages_mqtt,
)
from . import (
    constant,
    pipeline_stages_iothub,
    pipeline_events_iothub,
    pipeline_ops_iothub,
    pipeline_stages_iothub_mqtt,
)

logger = logging.getLogger(__name__)


class MQTTPipeline(object):
    def __init__(self, pipeline_configuration):
        """
        Constructor for instantiating a pipeline adapter object
        :param auth_provider: The authentication provider
        :param pipeline_configuration: The configuration generated based on user inputs
        """

        self.feature_enabled = {
            constant.C2D_MSG: False,
            constant.INPUT_MSG: False,
            constant.METHODS: False,
            constant.TWIN: False,
            constant.TWIN_PATCHES: False,
        }

        # Event Handlers - Will be set by Client after instantiation of this object
        self.on_connected = None
        self.on_disconnected = None
        self.on_c2d_message_received = None
        self.on_input_message_received = None
        self.on_method_request_received = None
        self.on_twin_patch_received = None

        # Currently a single timeout stage and a single retry stage for MQTT retry only.
        # Later, a higher level timeout and a higher level retry stage.
        self._pipeline = (
            #
            # The root is always the root.  By definition, it's the first stage in the pipeline.
            #
            pipeline_stages_base.PipelineRootStage(pipeline_configuration)
            #
            # SasTokenRenewalStage comes near the root by default because it should be as close
            # to the top of the pipeline as possible, and does not need to be after anything.
            #
            .append_stage(pipeline_stages_base.SasTokenRenewalStage())
            #
            # EnsureDesiredPropertiesStage needs to be above TwinRequestResponseStage because it
            # sends GetTwinOperation ops and that stage handles those ops.
            #
            .append_stage(pipeline_stages_iothub.EnsureDesiredPropertiesStage())
            #
            # TwinRequestResponseStage comes near the root by default because it doesn't need to be
            # after anything
            #
            .append_stage(pipeline_stages_iothub.TwinRequestResponseStage())
            #
            # CoordinateRequestAndResponseStage needs to be after TwinRequestResponseStage because
            # TwinRequestResponseStage creates the request ops that CoordinateRequestAndResponseStage
            # is coordinating.  It needs to be before IoTHubMQTTTranslationStage because that stage
            # operates on ops that CoordinateRequestAndResponseStage produces
            #
            .append_stage(pipeline_stages_base.CoordinateRequestAndResponseStage())
            #
            # IoTHubMQTTTranslationStage comes here because this is the point where we can translate
            # all operations directly into MQTT.  After this stage, only pipeline_stages_base stages
            # are allowed because IoTHubMQTTTranslationStage removes all the IoTHub-ness from the ops
            #
            .append_stage(pipeline_stages_iothub_mqtt.IoTHubMQTTTranslationStage())
            #
            # AutoConnectStage comes here because only MQTT ops have the need_connection flag set
            # and this is the first place in the pipeline wherer we can guaranetee that all network
            # ops are MQTT ops.
            #
            .append_stage(pipeline_stages_base.AutoConnectStage())
            #
            # ReconnectStage needs to be after AutoConnectStage because ReconnectStage sets/clears
            # the virtually_conencted flag and we want an automatic connection op to set this flag so
            # we can reconnect autoconnect operations.  This is important, for example, if a
            # send_message causes the transport to automatically connect, but that connection fails.
            # When that happens, the ReconenctState will hold onto the ConnectOperation until it
            # succeeds, and only then will return success to the AutoConnectStage which will
            # allow the publish to continue.
            #
            .append_stage(pipeline_stages_base.ReconnectStage())
            #
            # ConnectionLockStage needs to be after ReconnectStage because we want any ops that
            # ReconnectStage creates to go through the ConnectionLockStage gate
            #
            .append_stage(pipeline_stages_base.ConnectionLockStage())
            #
            # RetryStage needs to be near the end because it's retrying low-level MQTT operations.
            #
            .append_stage(pipeline_stages_base.RetryStage())
            #
            # OpTimeoutStage needs to be after RetryStage because OpTimeoutStage returns the timeout
            # errors that RetryStage is watching for.
            #
            .append_stage(pipeline_stages_base.OpTimeoutStage())
            #
            # MQTTTransportStage needs to be at the very end of the pipeline because this is where
            # operations turn into network traffic
            #
            .append_stage(pipeline_stages_mqtt.MQTTTransportStage())
        )

        def _on_pipeline_event(event):
            if isinstance(event, pipeline_events_iothub.C2DMessageEvent):
                if self.on_c2d_message_received:
                    self.on_c2d_message_received(event.message)
                else:
                    logger.warning("C2D message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.InputMessageEvent):
                if self.on_input_message_received:
                    self.on_input_message_received(event.input_name, event.message)
                else:
                    logger.warning("input message event received with no handler.  dropping.")

            elif isinstance(event, pipeline_events_iothub.MethodRequestEvent):
                if self.on_method_request_received:
                    self.on_method_request_received(event.method_request)
                else:
                    logger.warning("Method request event received with no handler. Dropping.")

            elif isinstance(event, pipeline_events_iothub.TwinDesiredPropertiesPatchEvent):
                if self.on_twin_patch_received:
                    self.on_twin_patch_received(event.patch)
                else:
                    logger.warning("Twin patch event received with no handler. Dropping.")

            else:
                logger.warning("Dropping unknown pipeline event {}".format(event.name))

        def _on_connected():
            if self.on_connected:
                self.on_connected()

        def _on_disconnected():
            if self.on_disconnected:
                self.on_disconnected()

        self._pipeline.on_pipeline_event_handler = _on_pipeline_event
        self._pipeline.on_connected_handler = _on_connected
        self._pipeline.on_disconnected_handler = _on_disconnected

        callback = EventedCallback()

        # NOTE: It would be nice if this didn't have to go down as a dynamic operation.
        # At this time we haven't been able to figure out a better way to make it work though.

        op = pipeline_ops_base.InitializePipelineOperation(callback=callback)

        self._pipeline.run_op(op)
        callback.wait_for_completion()

    def connect(self, callback):
        """
        Connect to the service.

        :param callback: callback which is called when the connection to the service is complete.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("Starting ConnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.ConnectOperation(callback=on_complete))

    def disconnect(self, callback):
        """
        Disconnect from the service.

        :param callback: callback which is called when the connection to the service has been disconnected

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("Starting DisconnectOperation on the pipeline")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(pipeline_ops_base.DisconnectOperation(callback=on_complete))

    def send_message(self, message, callback):
        """
        Send a telemetry message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendD2CMessageOperation(message=message, callback=on_complete)
        )

    def send_output_message(self, message, callback):
        """
        Send an output message to the service.

        :param message: message to send.
        :param callback: callback which is called when the message publish has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendOutputMessageOperation(message=message, callback=on_complete)
        )

    def send_method_response(self, method_response, callback):
        """
        Send a method response to the service.

        :param method_response: the method response to send
        :param callback: callback which is called when response has been acknowledged by the service

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """
        logger.debug("MQTTPipeline send_method_response called")

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.SendMethodResponseOperation(
                method_response=method_response, callback=on_complete
            )
        )

    def get_twin(self, callback):
        """
        Send a request for a full twin to the service.

        :param callback: callback which is called when request has been acknowledged by the service.
        This callback should have two parameters.  On success, this callback is called with the
        requested twin and error=None.  On failure, this callback is called with None for the requested
        twin and error set to the cause of the failure.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            if error:
                callback(error=error, twin=None)
            else:
                callback(twin=op.twin)

        self._pipeline.run_op(pipeline_ops_iothub.GetTwinOperation(callback=on_complete))

    def patch_twin_reported_properties(self, patch, callback):
        """
        Send a patch for a twin's reported properties to the service.

        :param patch: the reported properties patch to send
        :param callback: callback which is called when request has been acknowledged by the service.

        The following exceptions are not "raised", but rather returned via the "error" parameter
        when invoking "callback":

        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionFailedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ConnectionDroppedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.UnauthorizedError`
        :raises: :class:`azure.iot.device.iothub.pipeline.exceptions.ProtocolClientError`
        """

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_iothub.PatchTwinReportedPropertiesOperation(
                patch=patch, callback=on_complete
            )
        )

    def enable_feature(self, feature_name, callback):
        """
        Enable the given feature by subscribing to the appropriate topics.

        :param feature_name: one of the feature name constants from constant.py
        :param callback: callback which is called when the feature is enabled

        :raises: ValueError if feature_name is invalid
        """
        logger.debug("enable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")

        def on_complete(op, error):
            if error:
                logger.warning(
                    "Subscribe for {} failed.  Not enabling feature".format(feature_name)
                )
            else:
                self.feature_enabled[feature_name] = True
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.EnableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )

    def disable_feature(self, feature_name, callback):
        """
        Disable the given feature by subscribing to the appropriate topics.
        :param callback: callback which is called when the feature is disabled

        :param feature_name: one of the feature name constants from constant.py

        :raises: ValueError if feature_name is invalid
        """
        logger.debug("disable_feature {} called".format(feature_name))
        if feature_name not in self.feature_enabled:
            raise ValueError("Invalid feature_name")
        self.feature_enabled[feature_name] = False

        def on_complete(op, error):
            callback(error=error)

        self._pipeline.run_op(
            pipeline_ops_base.DisableFeatureOperation(
                feature_name=feature_name, callback=on_complete
            )
        )

    @property
    def connected(self):
        """
        Read-only property to indicate if the transport is connected or not.
        """
        return self._pipeline.connected



"""Azure IoT Device Common

This package provides shared modules for use with various Azure IoT device-side clients.

INTERNAL USAGE ONLY
"""

from .models import X509, ProxyOptions

__all__ = ["X509", "ProxyOptions"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from . import PipelineEvent


class IncomingMQTTMessageEvent(PipelineEvent):
    """
    A PipelineEvent object which represents an incoming MQTT message on some MQTT topic
    """

    def __init__(self, topic, payload):
        """
        Initializer for IncomingMQTTMessageEvent objects.

        :param str topic: The name of the topic that the incoming message arrived on.
        :param str payload: The payload of the message
        """
        super(IncomingMQTTMessageEvent, self).__init__()
        self.topic = topic
        self.payload = payload



"""Azure IoT Hub Device SDK Pipeline

This package provides a protocol pipeline for use with the Azure IoT Hub Device SDK.

INTERNAL USAGE ONLY
"""

from .mqtt_pipeline import MQTTPipeline
from .http_pipeline import HTTPPipeline
from .config import IoTHubPipelineConfig



""" Azure IoT Device Library

This library provides clients and associated models for communicating with Azure IoT services
from an IoT device.
"""

# Import all exposed items in subpackages to expose them via this package
from .iothub import *
from .provisioning import *
from .common import *  # TODO: do we really want to do this?

# Import the subpackages themselves in order to set the __all__
from . import iothub
from . import provisioning
from . import common

# Import the module to generate missing documentation
from . import patch_documentation


# TODO: remove this chunk of commented code if we truly no longer want to take this approach

# if sys.version_info > (3, 5):  # This only works for python 3.5+ at present
# Dynamically patch the clients to add shim implementations for all the inherited methods.
# This is necessary to generate accurate online docs.
# It SHOULD not impact the functionality of the methods themselves in any way.

# NOTE In the event of addition of new methods and generation of accurate documentation
# for those methods we have to append content to "patch_documentation.py" file.
# In order to do so please uncomment the "patch.add_shims" lines below,
# enable logging with level "DEBUG" in a python terminal and do
# "import azure.iot.device". The delta between the newly generated output
# and the existing content of "patch_documentation.py" should be appended to
# the function "execute_patch_for_sync" in "patch_documentation.py".
# Once done please again omment out the "patch.add_shims" lines below.

# patch.add_shims_for_inherited_methods(IoTHubDeviceClient)  # noqa: F405
# patch.add_shims_for_inherited_methods(IoTHubModuleClient)  # noqa: F405
# patch.add_shims_for_inherited_methods(ProvisioningDeviceClient)  # noqa: F405
patch_documentation.execute_patch_for_sync()


# iothub and common subpackages are still showing up in intellisense

__all__ = iothub.__all__ + provisioning.__all__ + common.__all__



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six.moves.urllib as urllib

logger = logging.getLogger(__name__)


def get_method_invoke_path(device_id, module_id=None):
    """
    :return: The path for invoking methods from one module to a device or module. It is of the format
    twins/uri_encode($device_id)/modules/uri_encode($module_id)/methods
    """
    if module_id:
        return "twins/{device_id}/modules/{module_id}/methods".format(
            device_id=urllib.parse.quote_plus(device_id),
            module_id=urllib.parse.quote_plus(module_id),
        )
    else:
        return "twins/{device_id}/methods".format(device_id=urllib.parse.quote_plus(device_id))


def get_storage_info_for_blob_path(device_id):
    """
    This does not take a module_id since get_storage_info_for_blob_path should only ever be invoked on device clients.

    :return: The path for getting the storage sdk credential information from IoT Hub. It is of the format
    devices/uri_encode($device_id)/files
    """
    return "devices/{}/files".format(urllib.parse.quote_plus(device_id))


def get_notify_blob_upload_status_path(device_id):
    """
    This does not take a module_id since get_notify_blob_upload_status_path should only ever be invoked on device clients.

    :return: The path for getting the storage sdk credential information from IoT Hub. It is of the format
    devices/uri_encode($device_id)/files/notifications
    """
    return "devices/{}/files/notifications".format(urllib.parse.quote_plus(device_id))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module provides hard coded patches used to modify items from the libraries.
Currently we have to do like this so that we don't use exec anywhere"""


def execute_patch_for_async():
    from azure.iot.device.iothub.aio.async_clients import IoTHubDeviceClient as IoTHubDeviceClient_

    async def connect(self):
        return await super(IoTHubDeviceClient_, self).connect()

    connect.__doc__ = IoTHubDeviceClient_.connect.__doc__
    setattr(IoTHubDeviceClient_, "connect", connect)

    async def disconnect(self):
        return await super(IoTHubDeviceClient_, self).disconnect()

    disconnect.__doc__ = IoTHubDeviceClient_.disconnect.__doc__
    setattr(IoTHubDeviceClient_, "disconnect", disconnect)

    async def get_storage_info_for_blob(self, blob_name):
        return await super(IoTHubDeviceClient_, self).get_storage_info_for_blob(blob_name)

    get_storage_info_for_blob.__doc__ = IoTHubDeviceClient_.get_storage_info_for_blob.__doc__
    setattr(IoTHubDeviceClient_, "get_storage_info_for_blob", get_storage_info_for_blob)

    async def get_twin(self):
        return await super(IoTHubDeviceClient_, self).get_twin()

    get_twin.__doc__ = IoTHubDeviceClient_.get_twin.__doc__
    setattr(IoTHubDeviceClient_, "get_twin", get_twin)

    async def notify_blob_upload_status(
        self, correlation_id, is_success, status_code, status_description
    ):
        return await super(IoTHubDeviceClient_, self).notify_blob_upload_status(
            correlation_id, is_success, status_code, status_description
        )

    notify_blob_upload_status.__doc__ = IoTHubDeviceClient_.notify_blob_upload_status.__doc__
    setattr(IoTHubDeviceClient_, "notify_blob_upload_status", notify_blob_upload_status)

    async def patch_twin_reported_properties(self, reported_properties_patch):
        return await super(IoTHubDeviceClient_, self).patch_twin_reported_properties(
            reported_properties_patch
        )

    patch_twin_reported_properties.__doc__ = (
        IoTHubDeviceClient_.patch_twin_reported_properties.__doc__
    )
    setattr(IoTHubDeviceClient_, "patch_twin_reported_properties", patch_twin_reported_properties)

    async def receive_method_request(self, method_name=None):
        return await super(IoTHubDeviceClient_, self).receive_method_request(method_name)

    receive_method_request.__doc__ = IoTHubDeviceClient_.receive_method_request.__doc__
    setattr(IoTHubDeviceClient_, "receive_method_request", receive_method_request)

    async def receive_twin_desired_properties_patch(self):
        return await super(IoTHubDeviceClient_, self).receive_twin_desired_properties_patch()

    receive_twin_desired_properties_patch.__doc__ = (
        IoTHubDeviceClient_.receive_twin_desired_properties_patch.__doc__
    )
    setattr(
        IoTHubDeviceClient_,
        "receive_twin_desired_properties_patch",
        receive_twin_desired_properties_patch,
    )

    async def send_message(self, message):
        return await super(IoTHubDeviceClient_, self).send_message(message)

    send_message.__doc__ = IoTHubDeviceClient_.send_message.__doc__
    setattr(IoTHubDeviceClient_, "send_message", send_message)

    async def send_method_response(self, method_response):
        return await super(IoTHubDeviceClient_, self).send_method_response(method_response)

    send_method_response.__doc__ = IoTHubDeviceClient_.send_method_response.__doc__
    setattr(IoTHubDeviceClient_, "send_method_response", send_method_response)

    def create_from_connection_string(cls, connection_string, **kwargs):
        return super(IoTHubDeviceClient_, cls).create_from_connection_string(
            connection_string, **kwargs
        )

    create_from_connection_string.__doc__ = (
        IoTHubDeviceClient_.create_from_connection_string.__doc__
    )
    setattr(
        IoTHubDeviceClient_,
        "create_from_connection_string",
        classmethod(create_from_connection_string),
    )

    def create_from_symmetric_key(cls, symmetric_key, hostname, device_id, **kwargs):
        return super(IoTHubDeviceClient_, cls).create_from_symmetric_key(
            symmetric_key, hostname, device_id, **kwargs
        )

    create_from_symmetric_key.__doc__ = IoTHubDeviceClient_.create_from_symmetric_key.__doc__
    setattr(
        IoTHubDeviceClient_, "create_from_symmetric_key", classmethod(create_from_symmetric_key)
    )

    def create_from_x509_certificate(cls, x509, hostname, device_id, **kwargs):
        return super(IoTHubDeviceClient_, cls).create_from_x509_certificate(
            x509, hostname, device_id, **kwargs
        )

    create_from_x509_certificate.__doc__ = IoTHubDeviceClient_.create_from_x509_certificate.__doc__
    setattr(
        IoTHubDeviceClient_,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )
    from azure.iot.device.iothub.aio.async_clients import IoTHubModuleClient as IoTHubModuleClient_

    async def connect(self):
        return await super(IoTHubModuleClient_, self).connect()

    connect.__doc__ = IoTHubModuleClient_.connect.__doc__
    setattr(IoTHubModuleClient_, "connect", connect)

    async def disconnect(self):
        return await super(IoTHubModuleClient_, self).disconnect()

    disconnect.__doc__ = IoTHubModuleClient_.disconnect.__doc__
    setattr(IoTHubModuleClient_, "disconnect", disconnect)

    async def get_storage_info_for_blob(self, blob_name):
        return await super(IoTHubModuleClient_, self).get_storage_info_for_blob(blob_name)

    get_storage_info_for_blob.__doc__ = IoTHubModuleClient_.get_storage_info_for_blob.__doc__
    setattr(IoTHubModuleClient_, "get_storage_info_for_blob", get_storage_info_for_blob)

    async def get_twin(self):
        return await super(IoTHubModuleClient_, self).get_twin()

    get_twin.__doc__ = IoTHubModuleClient_.get_twin.__doc__
    setattr(IoTHubModuleClient_, "get_twin", get_twin)

    async def notify_blob_upload_status(
        self, correlation_id, is_success, status_code, status_description
    ):
        return await super(IoTHubModuleClient_, self).notify_blob_upload_status(
            correlation_id, is_success, status_code, status_description
        )

    notify_blob_upload_status.__doc__ = IoTHubModuleClient_.notify_blob_upload_status.__doc__
    setattr(IoTHubModuleClient_, "notify_blob_upload_status", notify_blob_upload_status)

    async def patch_twin_reported_properties(self, reported_properties_patch):
        return await super(IoTHubModuleClient_, self).patch_twin_reported_properties(
            reported_properties_patch
        )

    patch_twin_reported_properties.__doc__ = (
        IoTHubModuleClient_.patch_twin_reported_properties.__doc__
    )
    setattr(IoTHubModuleClient_, "patch_twin_reported_properties", patch_twin_reported_properties)

    async def receive_method_request(self, method_name=None):
        return await super(IoTHubModuleClient_, self).receive_method_request(method_name)

    receive_method_request.__doc__ = IoTHubModuleClient_.receive_method_request.__doc__
    setattr(IoTHubModuleClient_, "receive_method_request", receive_method_request)

    async def receive_twin_desired_properties_patch(self):
        return await super(IoTHubModuleClient_, self).receive_twin_desired_properties_patch()

    receive_twin_desired_properties_patch.__doc__ = (
        IoTHubModuleClient_.receive_twin_desired_properties_patch.__doc__
    )
    setattr(
        IoTHubModuleClient_,
        "receive_twin_desired_properties_patch",
        receive_twin_desired_properties_patch,
    )

    async def send_message(self, message):
        return await super(IoTHubModuleClient_, self).send_message(message)

    send_message.__doc__ = IoTHubModuleClient_.send_message.__doc__
    setattr(IoTHubModuleClient_, "send_message", send_message)

    async def send_method_response(self, method_response):
        return await super(IoTHubModuleClient_, self).send_method_response(method_response)

    send_method_response.__doc__ = IoTHubModuleClient_.send_method_response.__doc__
    setattr(IoTHubModuleClient_, "send_method_response", send_method_response)

    def create_from_connection_string(cls, connection_string, **kwargs):
        return super(IoTHubModuleClient_, cls).create_from_connection_string(
            connection_string, **kwargs
        )

    create_from_connection_string.__doc__ = (
        IoTHubModuleClient_.create_from_connection_string.__doc__
    )
    setattr(
        IoTHubModuleClient_,
        "create_from_connection_string",
        classmethod(create_from_connection_string),
    )

    def create_from_edge_environment(cls, **kwargs):
        return super(IoTHubModuleClient_, cls).create_from_edge_environment(**kwargs)

    create_from_edge_environment.__doc__ = IoTHubModuleClient_.create_from_edge_environment.__doc__
    setattr(
        IoTHubModuleClient_,
        "create_from_edge_environment",
        classmethod(create_from_edge_environment),
    )

    def create_from_x509_certificate(cls, x509, hostname, device_id, module_id, **kwargs):
        return super(IoTHubModuleClient_, cls).create_from_x509_certificate(
            x509, hostname, device_id, module_id, **kwargs
        )

    create_from_x509_certificate.__doc__ = IoTHubModuleClient_.create_from_x509_certificate.__doc__
    setattr(
        IoTHubModuleClient_,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )
    from azure.iot.device.provisioning.aio.async_provisioning_device_client import (
        ProvisioningDeviceClient as ProvisioningDeviceClient_,
    )

    def create_from_symmetric_key(
        cls, provisioning_host, registration_id, id_scope, symmetric_key, **kwargs
    ):
        return super(ProvisioningDeviceClient_, cls).create_from_symmetric_key(
            provisioning_host, registration_id, id_scope, symmetric_key, **kwargs
        )

    create_from_symmetric_key.__doc__ = ProvisioningDeviceClient_.create_from_symmetric_key.__doc__
    setattr(
        ProvisioningDeviceClient_,
        "create_from_symmetric_key",
        classmethod(create_from_symmetric_key),
    )

    def create_from_x509_certificate(
        cls, provisioning_host, registration_id, id_scope, x509, **kwargs
    ):
        return super(ProvisioningDeviceClient_, cls).create_from_x509_certificate(
            provisioning_host, registration_id, id_scope, x509, **kwargs
        )

    create_from_x509_certificate.__doc__ = (
        ProvisioningDeviceClient_.create_from_x509_certificate.__doc__
    )
    setattr(
        ProvisioningDeviceClient_,
        "create_from_x509_certificate",
        classmethod(create_from_x509_certificate),
    )



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six.moves.urllib as urllib

logger = logging.getLogger(__name__)

# NOTE: Whenever using standard URL encoding via the urllib.parse.quote() API
# make sure to specify that there are NO safe values (e.g. safe=""). By default
# "/" is skipped in encoding, and that is not desirable.
#
# DO NOT use urllib.parse.quote_plus(), as it turns ' ' characters into '+',
# which is invalid for MQTT publishes.
#
# DO NOT use urllib.parse.unquote_plus(), as it turns '+' characters into ' ',
# which is also invalid.


def _get_topic_base():
    """
    return the string that creates the beginning of all topics for DPS
    """
    return "$dps/registrations/"


def get_register_topic_for_subscribe():
    """
    :return: The topic string used to subscribe for receiving future responses from DPS.
    It is of the format "$dps/registrations/res/#"
    """
    return _get_topic_base() + "res/#"


def get_register_topic_for_publish(request_id):
    """
    :return: The topic string used to send a registration. It is of the format
    "$dps/registrations/PUT/iotdps-register/?$rid=<request_id>
    """
    return (_get_topic_base() + "PUT/iotdps-register/?$rid={request_id}").format(
        request_id=urllib.parse.quote(str(request_id), safe="")
    )


def get_query_topic_for_publish(request_id, operation_id):
    """
    :return: The topic string used to send a query. It is of the format
    "$dps/registrations/GET/iotdps-get-operationstatus/?$rid=<request_id>&operationId=<operation_id>
    """
    return (
        _get_topic_base()
        + "GET/iotdps-get-operationstatus/?$rid={request_id}&operationId={operation_id}"
    ).format(
        request_id=urllib.parse.quote(str(request_id), safe=""),
        operation_id=urllib.parse.quote(str(operation_id), safe=""),
    )


def _get_topic_for_response():
    """
    return the topic string used to publish telemetry
    """
    return _get_topic_base() + "res/"


def is_dps_response_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>...&<keyN>=<valueN>
    :param topic: The topic string
    """
    if _get_topic_for_response() in topic:
        return True
    return False


def extract_properties_from_dps_response_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>...&<keyN>=<valueN>
    Extract key=value pairs from the latter part of the topic.
    :param topic: The topic string
    :return: a dictionary of property keys mapped to property values.
    """
    topic_parts = topic.split("$")
    properties = topic_parts[2]

    # NOTE: we cannot use urllib.parse.parse_qs because it always decodes '+' as ' ',
    # and the behavior cannot be overriden. Must parse key/value pairs manually.

    if properties:
        key_value_pairs = properties.split("&")
        key_value_dict = {}
        for entry in key_value_pairs:
            pair = entry.split("=")
            key = urllib.parse.unquote(pair[0])
            value = urllib.parse.unquote(pair[1])
            if key_value_dict.get(key):
                raise ValueError("Duplicate keys in DPS response topic")
            else:
                key_value_dict[key] = value

    return key_value_dict


def extract_status_code_from_dps_response_topic(topic):
    """
    Topics for responses from DPS are of the following format:
    $dps/registrations/res/<statuscode>/?$<key1>=<value1>&<key2>=<value2>...&<keyN>=<valueN>
    Extract the status code part from the topic.
    :param topic: The topic string
    :return: The status code from the DPS response topic, as a string
    """
    POS_STATUS_CODE_IN_TOPIC = 3
    topic_parts = topic.split("$")
    url_parts = topic_parts[1].split("/")
    status_code = url_parts[POS_STATUS_CODE_IN_TOPIC]
    return urllib.parse.unquote(status_code)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains tools for working with Connection Strings"""

__all__ = ["ConnectionString"]

CS_DELIMITER = ";"
CS_VAL_SEPARATOR = "="

HOST_NAME = "HostName"
SHARED_ACCESS_KEY_NAME = "SharedAccessKeyName"
SHARED_ACCESS_KEY = "SharedAccessKey"
SHARED_ACCESS_SIGNATURE = "SharedAccessSignature"
DEVICE_ID = "DeviceId"
MODULE_ID = "ModuleId"
GATEWAY_HOST_NAME = "GatewayHostName"

_valid_keys = [
    HOST_NAME,
    SHARED_ACCESS_KEY_NAME,
    SHARED_ACCESS_KEY,
    SHARED_ACCESS_SIGNATURE,
    DEVICE_ID,
    MODULE_ID,
    GATEWAY_HOST_NAME,
]


def _parse_connection_string(connection_string):
    """Return a dictionary of values contained in a given connection string
    """
    try:
        cs_args = connection_string.split(CS_DELIMITER)
    except (AttributeError, TypeError):
        # NOTE: in Python 2.7, bytes will not raise an error here as they do in all other versions
        raise TypeError("Connection String must be of type str")
    try:
        d = dict(arg.split(CS_VAL_SEPARATOR, 1) for arg in cs_args)
    except ValueError:
        # This occurs in an extreme edge case where a dictionary cannot be formed because there
        # is only 1 token after the split (dict requires two in order to make a key/value pair)
        raise ValueError("Invalid Connection String - Unable to parse")
    if len(cs_args) != len(d):
        # various errors related to incorrect parsing - duplicate args, bad syntax, etc.
        raise ValueError("Invalid Connection String - Unable to parse")
    if not all(key in _valid_keys for key in d.keys()):
        raise ValueError("Invalid Connection String - Invalid Key")
    _validate_keys(d)
    return d


def _validate_keys(d):
    """Raise ValueError if incorrect combination of keys in dict d
    """
    host_name = d.get(HOST_NAME)
    shared_access_key_name = d.get(SHARED_ACCESS_KEY_NAME)
    shared_access_key = d.get(SHARED_ACCESS_KEY)
    device_id = d.get(DEVICE_ID)

    # This logic could be expanded to return the category of ConnectionString
    if host_name and device_id and shared_access_key:
        pass
    elif host_name and shared_access_key and shared_access_key_name:
        pass
    else:
        raise ValueError("Invalid Connection String - Incomplete")


class ConnectionString(object):
    """Key/value mappings for connection details.
    Uses the same syntax as dictionary
    """

    def __init__(self, connection_string):
        """Initializer for ConnectionString

        :param str connection_string: String with connection details provided by Azure
        :raises: ValueError if provided connection_string is invalid
        """
        self._dict = _parse_connection_string(connection_string)
        self._strrep = connection_string

    def __getitem__(self, key):
        return self._dict[key]

    def __repr__(self):
        return self._strrep

    def get(self, key, default=None):
        """Return the value for key if key is in the dictionary, else default

        :param str key: The key to retrieve a value for
        :param str default: The default value returned if a key is not found
        :returns: The value for the given key
        """
        try:
            return self._dict[key]
        except KeyError:
            return default



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import logging
import six
import traceback
import copy
from . import (
    pipeline_ops_base,
    PipelineStage,
    pipeline_ops_http,
    pipeline_thread,
    pipeline_exceptions,
)
from azure.iot.device.common.http_transport import HTTPTransport
from azure.iot.device.common import handle_exceptions, transport_exceptions
from azure.iot.device.common.callable_weak_method import CallableWeakMethod

logger = logging.getLogger(__name__)


class HTTPTransportStage(PipelineStage):
    """
    PipelineStage object which is responsible for interfacing with the HTTP protocol wrapper object.
    This stage handles all HTTP operations that are not specific to IoT Hub.
    """

    def __init__(self):
        super(HTTPTransportStage, self).__init__()
        # The sas_token will be set when Connetion Args are received
        self.sas_token = None

        # The transport will be instantiated when Connection Args are received
        self.transport = None

    @pipeline_thread.runs_on_pipeline_thread
    def _run_op(self, op):
        if isinstance(op, pipeline_ops_base.InitializePipelineOperation):

            # If there is a gateway hostname, use that as the hostname for connection,
            # rather than the hostname itself
            if self.pipeline_root.pipeline_configuration.gateway_hostname:
                logger.debug(
                    "Gateway Hostname Present. Setting Hostname to: {}".format(
                        self.pipeline_root.pipeline_configuration.gateway_hostname
                    )
                )
                hostname = self.pipeline_root.pipeline_configuration.gateway_hostname
            else:
                logger.debug(
                    "Gateway Hostname not present. Setting Hostname to: {}".format(
                        self.pipeline_root.pipeline_configuration.hostname
                    )
                )
                hostname = self.pipeline_root.pipeline_configuration.hostname

            # Create HTTP Transport
            logger.debug("{}({}): got connection args".format(self.name, op.name))
            self.transport = HTTPTransport(
                hostname=hostname,
                server_verification_cert=self.pipeline_root.pipeline_configuration.server_verification_cert,
                x509_cert=self.pipeline_root.pipeline_configuration.x509,
                cipher=self.pipeline_root.pipeline_configuration.cipher,
            )

            self.pipeline_root.transport = self.transport
            op.complete()

        elif isinstance(op, pipeline_ops_http.HTTPRequestAndResponseOperation):
            # This will call down to the HTTP Transport with a request and also created a request callback. Because the HTTP Transport will run on the http transport thread, this call should be non-blocking to the pipline thread.
            logger.debug(
                "{}({}): Generating HTTP request and setting callback before completing.".format(
                    self.name, op.name
                )
            )

            @pipeline_thread.invoke_on_pipeline_thread_nowait
            def on_request_completed(error=None, response=None):
                if error:
                    logger.error(
                        "{}({}): Error passed to on_request_completed. Error={}".format(
                            self.name, op.name, error
                        )
                    )
                    op.complete(error=error)
                else:
                    logger.debug(
                        "{}({}): Request completed. Completing op.".format(self.name, op.name)
                    )
                    logger.debug("HTTP Response Status: {}".format(response["status_code"]))
                    logger.debug("HTTP Response: {}".format(response["resp"].decode("utf-8")))
                    op.response_body = response["resp"]
                    op.status_code = response["status_code"]
                    op.reason = response["reason"]
                    op.complete()

            # A deepcopy is necessary here since otherwise the manipulation happening to
            # http_headers will affect the op.headers, which would be an unintended side effect
            # and not a good practice.
            http_headers = copy.deepcopy(op.headers)
            if self.pipeline_root.pipeline_configuration.sastoken:
                http_headers["Authorization"] = str(
                    self.pipeline_root.pipeline_configuration.sastoken
                )

            self.transport.request(
                method=op.method,
                path=op.path,
                headers=http_headers,
                query_params=op.query_params,
                body=op.body,
                callback=on_request_completed,
            )

        else:
            self.send_op_down(op)
